<?php
/**
 * Datasource Helper Plugin.
 *
 * @license  Apache 2 (http://www.apache.org/licenses/LICENSE-2.0)
 * @author   Kazuyuki Matsuda / kazm.atmark.shorindo.com
 */
if (!defined('DOKU_INC')) die();
if (!defined('DOKU_LF')) define('DOKU_LF', "\n");
if (!defined('DOKU_TAB')) define('DOKU_TAB', "\t");

class helper_plugin_datasource extends DokuWiki_Plugin {
    var $_dsn;

    function __construct() {
        $this->_dsn = array();
        $lines = preg_split('/\s+/', $this->getConf('dsn'));
        foreach ($lines as $line) {
            if (preg_match('/^([^=]+)=([^\|]+)(\|([^\|]+))?(\|([^\|]+))?$/', $line, $m)) {
                $this->_dsn[$m[1]] = array($m[2], $m[4], $m[6]);
            }
        }
    }

	function getInfo() {
		return confToHash(dirname(__FILE__).'/plugin.info.txt');
	}

    function getMethods() {
        $result = array();
        $result[] = array(
            'name'   => 'bind',
            'desc'   => 'bind to datasource object',
            'params' => array(
                'name' => 'string'),
            'return' => array('object' => 'object'),
            );
        return $result;
    }

    function connect() {
        try {
            return new PDO($this->getConf('dsn'),
                           $this->getConf('dbuser'),
                           conf_decodeString($this->getConf('dbpass')));
        } catch (Exception $ex) {
            dbglog($ex->getMessage());
        }
    }

    function read($dsn) {
        $dso = new DataSourceObject();
        $dso->getinfo($dsn);

        try {
            switch ($dso->schema) {
            case 'wiki':
                $dso = new WikiDataSource();
                return $dso->read($dsn);

            case 'tag':
                $dso = new TagDataSource();
                return $dso->read($dsn);

            case 'media':
                $dso = new MediaDataSource();
                return $dso->read($dsn);

            default:
                if ($this->_dsn[$dso->schema]) {
                    $dso = new SqlDataSource($this->_dsn[$dso->schema][0],
                                             $this->_dsn[$dso->schema][1],
                                             $this->_dsn[$dso->schema][2]);
                    return $dso->read($dsn);
                } else {
                    $data = array('dsn' => $dsn);
                    $result = trigger_event('DATASOURCE_UNKNOWN_SCHEMA', $data);
                    //dbglog("helper_plugin_datasource::create(".print_r($data,true).")");
                    if ($data['result'])
                        return $data['result'];
                }
            }
        } catch (Exception $ex) {
            return $this->error("2", $ex->getMessage());
        }

        return $this->error('1', 'NO VALID DSN FOUND');
    }

    function error($code, $message) {
        $doc = new DOMDocument('1.0', 'UTF-8');
        $doc->preserveWhiteSpace = false;
        $doc->formatOutput = true;
        $result = $doc->appendChild($doc->createElement('result'));
        $result->appendChild($doc->createElement('code', $code));
        $result->appendChild($doc->createElement('message', $message));
        return $doc;
    }
}

/**
 *
 */
class DataSourceObject {
    var $dsn;
    var $schema;
    var $type;
    var $opts;

    function read($dsn) {
        $this->getinfo($dsn);
        $doc = new DOMDocument();
        $doc->preserveWhiteSpace = false;
        $doc->encoding = 'UTF-8';
        $doc->formatOutput = true;
        $dom = $doc->appendChild($doc->createElement('result'));
        $dom->setAttribute('dsn', $dsn);
        $this->_createnode($dom, 'code', 1);
        $this->_createnode($dom, 'message', 'INVALID DATASOURCE NAME');
        return $doc;
    }

    function getinfo($dsn) {
        $this->dsn = $dsn;
        $NS = getNS(getID());
        $p = explode('?', $dsn, 2);
        list($schema, $type) = explode(':', $p[0], 2);
        $opts = array();
        foreach (explode('&', $p[1]) as $o) {
            list($key, $val) = explode('=', $o, 2);
            $opts[$key] = preg_replace(array('/\$ID([^\d\w])?/u','/\$NS([^\d\w])?/u'),
                                       array($ID, $NS),
                                       $val);
        }
        $this->schema = $schema;
        $this->type = $type;
        $this->opts = $opts;
        return array('schema' => $schema, 'type' => $type, 'opts' => $opts);
    }

    function _createnode($dom, $name, $value = null) {
        $doc = $dom->ownerDocument;
        $node = $dom->appendChild($doc->createElement($name));
        if ($value !== null)
            $node->appendChild($doc->createTextNode($value));
        return $node;
    }

    function _nsmatch($id, $targets) {
        foreach ($targets as $target) {
            if ($id == $target)
                return true;
            if (strpos($id, $target.':', 0) === 0)
                return true;
        }
        return false;
    }

    function _gettitle($id) {
        if (useHeading('content') && $id) {
            $heading = p_get_first_heading($id, true);
            if ($heading) return $heading;
        }
        return noNS($id);
    }
    
}

/**
 * wiki:[ページID](?content|meta|index|changes)
 */
class WikiDataSource extends DataSourceObject {
    function read($dsn) {
        $this->getinfo($dsn);
        $id = $this->opts['id'] ? $this->opts['id'] : getID();
        $doc = new DOMDocument();
        $doc->preserveWhiteSpace = false;
        $doc->encoding = 'UTF-8';
        $doc->formatOutput = true;
        $dom = $doc->appendChild($doc->createElement('result'));
        $dom->setAttribute('dsn', $this->dsn);

        switch ($this->type) {
        case 'meta':
            $this->_meta($dom);
            break;

        case 'index':
            $this->_index($dom);
            break;

        case 'changes':
            $this->_changes($dom);
            break;

        case 'revisions':
            $this->_revisions($dom);
            break;

        default:
            $this->_createnode($dom, 'code', '2');
            $this->_createnode($dom, 'message', 'INVALID NAME "'.$this->name.'"');
            break;
        }
        return $doc;
    }

    function _meta($result) {
        $_id = $this->opts['id'] ? $this->opts['id'] : getID();
        $_meta = p_get_metadata($_id, false, true);
        if (!$_meta) {
            $this->_createnode($result, 'code', '1');
            $this->_createnode($result, 'message', 'NOT FOUND');
            $this->_createnode($result, 'id', $_id);
            return $result->ownerDocument;
        }

        $this->_createnode($result, 'code', '0');
        $this->_createnode($result, 'message', 'OK');
        $this->_createnode($result, 'id', $_id);
        $meta = $this->_createnode($result, 'meta');
        $date = $this->_createnode($meta, 'date');
        $this->_createnode($date, 'created', $_meta['date']['created']);
        $this->_createnode($date, 'modified', $_meta['date']['modified']);
        $this->_createnode($meta, 'creator', $_meta['creator']);
        $last_change = $this->_createnode($meta, 'last_change');
        $this->_createnode($last_change, 'date', $_meta['last_change']['date']);
        $this->_createnode($last_change, 'ip', $_meta['last_change']['ip']);
        $this->_createnode($last_change, 'type', $_meta['last_change']['type']);
        $this->_createnode($last_change, 'id', $_meta['last_change']['id']);
        $this->_createnode($last_change, 'user', $_meta['last_change']['user']);
        $this->_createnode($last_change, 'sum', $_meta['last_change']['sum']);
        $this->_createnode($last_change, 'extra', $_meta['last_change']['extra']);
        $description = $this->_createnode($meta, 'description');
        $toc = $this->_createnode($description, 'tableofcontents');
        $_toc = $_meta['description']['tableofcontents'];

        for ($i = 0; $i < count($_toc); $i++) {
            $dummy = array();
            $item = $this->_createnode($toc, 'item');
            $this->_createnode($item, 'link', '#'.sectionID($_toc[$i]['title'], $dummy));
            $this->_createnode($item, 'title', $_toc[$i]['title']);
            $this->_createnode($item, 'type', $_toc[$i]['type']);
            $this->_createnode($item, 'level', $_toc[$i]['level']);
        }

        $this->_createnode($description, 'abstract', $_meta['description']['abstract']);
        $internal = $this->_createnode($meta, 'internal');
        $this->_createnode($internal, 'cache', $_meta['internal']['cache']);
        $this->_createnode($internal, 'toc', $_meta['internal']['toc']);
        $relation = $this->_createnode($meta, 'relation');
        $this->_createnode($relation, 'toc', $_meta['relation']['firstimage']);

        return $result->ownerDocument;
    }

    function _index($result) {
        global $conf;
        $_ns  = $this->opts['ns'] ?
            $this->opts['ns'] :
            getNS(getID());

        $_nsenc  = utf8_encodeFN(str_replace(':','/',$_ns));
        $data = array();
        search($data, $conf['datadir'], 'search_index', array('ns' => $_nsenc, 'meta' => 1), $_nsenc, 1); 
        $this->_createnode($result, 'code', '0');
        $this->_createnode($result, 'message', 'OK');
        $this->_createnode($result, 'namespace', $_ns);

        if ($_ns != ':') {
            $parent = $this->_createnode($result, 'parent');
            $pid = tpl_getparent($_ns);
            $this->_createnode($parent, 'id', $pid);
            $this->_createnode($parent, 'title', $this->_gettitle($pid));
            $this->_createnode($parent, 'url', wl($pid));
        }

        //sort
        $pages = $this->_createnode($result, 'pages');
        $ex = explode(',', $this->opts['excludes']);
        $sorted = array();
        foreach ($data as $item) {
            //skip data specified 'excludes' parameter.
            if (count($ex) > 0 && $this->_nsmatch($item['id'], $ex))
                continue;
            $item['title'] = $this->_gettitle($item['id']);
            $sorted[] = $item;
        }

        switch ($this->opts['sort']) {
        case 'id': // by id asc
            usort($sorted, array($this, '_cmpid'));
            break;
        case 'title': //by title asc
            usort($sorted, array($this, '_cmptitle'));
	        break;
        case 'titledesc': //by title desc
            usort($sorted, array($this, '_cmptitledesc'));
            break;
        case 'mtime':
            usort($sorted, array($this, '_cmpmtime'));
            break;
        case 'mtimedesc':
            usort($sorted, array($this, '_cmpmtimedesc'));
            break;
        case 'iddesc': // by id desc
        default:
            usort($sorted, array($this, '_cmpiddesc'));
        }
            
        foreach ($sorted as $item) {
            $page = $this->_createnode($pages, 'page');
            $this->_createnode($page, 'title', $item['title']);
            $this->_createnode($page, 'id', $item['id']);
            $this->_createnode($page, 'type', $item['type']);
            $this->_createnode($page, 'level', $item['level']);
            $this->_createnode($page, 'open', $item['open']);
            $this->_createnode($page, 'url', wl($item['id']));
        }
    }

    function _cmptitle($a, $b) {
        //dbglog($a['title'].'<=>'.$b['title']);
        return strcmp($a['title'], $b['title']);
    }

    function _cmptitledesc($a, $b) {
        //dbglog($a['title'].'<=>'.$b['title']);
        return strcmp($b['title'], $a['title']);
    }

    function _cmpid($a, $b) {
        //dbglog($a['id'].'<=>'.$b['id']);
        return strcmp($a['id'], $b['id']);
    }

    function _cmpiddesc($a, $b) {
        //dbglog($a['id'].'<=>'.$b['id']);
        return strcmp($b['id'], $a['id']);
    }

    function _cmpmtime($a, $b) {
        dbglog($a['mtime'].'<=>'.$b['mtime']);
        return strcmp($a['mtime'], $b['mtime']);
    }

    function _cmpmtimedesc($a, $b) {
        dbglog($a['mtime'].'<=>'.$b['mtime']);
        return strcmp($b['mtime'], $a['mtime']);
    }

    function _changes($result) {
        $this->_createnode($result, 'code', '0');
        $this->_createnode($result, 'message', 'OK');
        $this->_createnode($result, 'namespace', $this->opts['ns']);
        $changes = $this->_createnode($result, 'changes');
        $types = $this->opts['type'] ?
            explode(',', $this->opts['type']) :
            array('C', 'E', 'R');        
        $ex = explode(',', $this->opts['excludes']);

        $limit = $this->opts['limit'] ? $this->opts['limit'] : 10;
        global $conf;
        $since = time() - $conf['recent_days'] * 24*60*60;
//      foreach (array_reverse(getRecentsSince($since, null, $this->opts['ns'], RECENTS_SKIP_DELETED)) as $item) {
      foreach (array_reverse(getRecents(0, $limit, $this->opts['ns'], RECENTS_SKIP_DELETED)) as $item) {
            //skip data specified 'type' parameter.
            if (!in_array($item['type'], $types))
                continue;

            //skip data specified 'excludes' parameter.
            if (count($ex) > 0 && $this->_nsmatch($item['id'], $ex))
                continue;

            if (--$limit < 0)
                break;

            $change = $this->_createnode($changes, 'change');
            $this->_createnode($change, 'title', $this->_gettitle($item['id']));
            foreach ($item as $key => $val) {
                if ($key == 'date') {
                    $p = $this->_createnode($change, $key);
                    $p->setAttribute('value', $val);
                    $d = explode(' ', @date('Y m d H i s', $val));
                    $p->setAttribute('Y', $d[0]);
                    $p->setAttribute('m', $d[1]);
                    $p->setAttribute('d', $d[2]);
                    $p->setAttribute('H', $d[3]);
                    $p->setAttribute('i', $d[4]);
                    $p->setAttribute('s', $d[5]);
                    $this->_createnode($p, 'date', $d[0].'-'.$d[1].'-'.$d[2]);
                    $this->_createnode($p, 'time', $d[3].':'.$d[4].':'.$d[5]);
                    continue;
                }
                $this->_createnode($change, $key, $val);
            }
            $this->_createnode($change, 'md_and_title',
                               date('m/d', $item['date']).' '.
                               $this->_gettitle($item['id']));
            $this->_createnode($change, 'url', wl($item['id']));
        }
    }

    function _revisions($result) {
        $this->opts['id'] = $this->opts['id'] ?
            $this->opts['id'] :
            getID();
        $this->_createnode($result, 'code', '0');
        $this->_createnode($result, 'message', 'OK');
        $this->_createnode($result, 'id', $this->opts['id']);
        $revisions = $this->_createnode($result, 'revisions');
        foreach (getRevisions($this->opts['id'], 0, 10) as $rev) {
            $info = getRevisionInfo($this->opts['id'], $rev);
            $revnode = $this->_createnode($revisions, 'revision');
            $revnode->setAttribute('rev', $rev);
            $datenode = $this->_createnode($revnode, 'date');
            $datenode->setAttribute('value', $info['date']);
            $d = explode(' ', @date('Y m d H i s', $rev));
            $datenode->setAttribute('Y', $d[0]);
            $datenode->setAttribute('m', $d[1]);
            $datenode->setAttribute('d', $d[2]);
            $datenode->setAttribute('H', $d[3]);
            $datenode->setAttribute('i', $d[4]);
            $datenode->setAttribute('s', $d[5]);
            $this->_createnode($datenode, 'date', $d[0].'-'.$d[1].'-'.$d[2]);
            $this->_createnode($datenode, 'time', $d[3].':'.$d[4].':'.$d[5]);
            $this->_createnode($revnode, 'ip', $info['ip']);
            $this->_createnode($revnode, 'type', $info['type']);
            $this->_createnode($revnode, 'id', $info['id']);
            $this->_createnode($revnode, 'user', $info['user']);
            $this->_createnode($revnode, 'sum', $info['sum']);
            $this->_createnode($revnode, 'extra', $info['extra']);
        }
    }
}

/**
 *
 */
class TagDataSource extends DataSourceObject {
    function read($dsn) {
        $this->getinfo($dsn);
        $doc = new DOMDocument();
        $doc->preserveWhiteSpace = false;
        $doc->encoding = 'UTF-8';
        $doc->formatOutput = true;
        $result = $doc->appendChild($doc->createElement('result'));
        $result->setAttribute('dsn', $this->dsn);

        $this->_createnode($result, 'code', 0);
        $this->_createnode($result, 'message', 'OK');
        switch($this->type) {
        case 'keys':
        case 'list':
            $this->_keys($result);
            break;

        case 'related':
            $this->_related($result);
            break;

        default:
        }
        return $doc;
    }

    function _keys(&$result) {
        $tagsnode = $this->_createnode($result, 'tags');
        $obj = plugin_load('helper', 'tag');
        if ($this->opts['sort'] == 'desc')
            krsort($obj->topic_idx);
        else
            ksort($obj->topic_idx);

        if ($this->opts['ns'] == '$REQUEST')
            $this->opts['ns'] = $_REQUEST['ns'];

        $ns = $this->opts['ns'] ? explode(',', $this->opts['ns']) : array();
        $ex = $this->opts['excludes'] ? explode(',', $this->opts['excludes']) : array();
        $tags = $this->opts['tags'] ? explode(',', $this->opts['tags']) : array();

        foreach ($obj->topic_idx as $key => $val) {
            //tagsを指定された場合、keyが一致しないものはスキップする
            if (count($tags) > 0 && !in_array($key, $tags)) {
                continue;
            }

            //nsを指定された場合、idがマッチしないものを取り除く
            $nsfiltered = array();
            foreach ($val as $pid) {
                if (count($ex) > 0 && $this->_nsmatch($pid, $ex))
                    continue;
                if (count($ns) == 0 || $this->_nsmatch($pid, $ns))
                    $nsfiltered[] = $pid;
            }

            //aclチェック
            $aclfiltered = array();
            foreach ($nsfiltered as $pid) {
                if (auth_quickaclcheck($pid) >= AUTH_READ)
                    $aclfiltered[] = $pid;
            }

            if (count($aclfiltered) == 0)
                continue;

            $tag = $this->_createnode($tagsnode, 'tag');
            $this->_createnode($tag, 'name', $key);
            $params = array('do' => 'showtag', 'tag' => $key);

            if ($this->opts['ns'])
                $params['ns'] = $this->opts['ns'];

            $this->_createnode($tag, 'link', wl('tag:'.$key, $params, false, '&'));
            $this->_createnode($tag, 'count', count($aclfiltered));

            if ($this->type == 'list') {
                $pages = $this->_createnode($tag, 'pages');

                foreach ($aclfiltered as $pid) {
                    $child = $this->_createnode($pages, 'page');
                    $this->_createnode($child, 'id', $pid);
                    $this->_createnode($child, 'title', $this->_gettitle($pid));
                    $this->_createnode($child, 'link', wl($pid));
                }
            }
        }

        return $doc;
    }

    function _related(&$result) {
        //$tagsnode = $this->_createnode($result, 'tags');
        $obj = plugin_load('helper', 'tag');
        if ($this->opts['sort'] == 'desc')
            krsort($obj->topic_idx);
        else
            ksort($obj->topic_idx);

        //現在のページのタグを取得
        $id = getID();
        $tags = array();
        foreach ($obj->topic_idx as $key => $val) {
            foreach ($val as $pid) {
                if ($pid == getID()) {
                    $tags[] = $key;
                    break;
                }
            }
        }

        //各タグに含まれるアクセス可能なページを取り出し、自分自身と重複を
        //取り除いてソートする。
        $pages = array();
        $ns = $this->opts['ns'] ? explode(',', $this->opts['ns']) : array();
        $ex = $this->opts['excludes'] ? explode(',', $this->opts['excludes']) : array();
        foreach ($tags as $tag) {
            foreach ($obj->topic_idx[$tag] as $page) {
                if (count($ex) > 0 && $this->_nsmatch($page, $ex))
                    continue;
                if ((count($ns) == 0 || $this->_nsmatch($page, $ns)) &&
                    !in_array($page, $pages) && $page != getID())
                    $pages[] = $page;
            }
        }
        $aclfiltered = array();
        foreach ($pages as $page) {
            //aclチェック
            if (auth_quickaclcheck($page) >= AUTH_READ)
                $aclfiltered[] = array('id'=>$page, 'link'=>wl($page), 'title'=>$this->_gettitle($page));
        }
        usort($aclfiltered, array($this, '_titlesort'));

        //DOM要素作成
        $pagesnode = $this->_createnode($result, 'pages');
        foreach ($aclfiltered as $page) {
            $pagenode = $this->_createnode($pagesnode, 'page');
            $this->_createnode($pagenode, 'id', $page['pid']);
            $this->_createnode($pagenode, 'link', $page['link']);
            $this->_createnode($pagenode, 'title', $page['title']);
        }
    }

    function _titlesort($a, $b) {
        return strcmp($a['title'], $b['title']);
    }
}

/**
 *
 */
class SqlDataSource extends DataSourceObject {
    var $_conn;

    function __construct($dbname, $dbuser, $dbpass) {
        $this->_conn = new PDO($dbname, $dbuser, $dbpass, array(
            PDO::MYSQL_ATTR_INIT_COMMAND => "SET CHARACTER SET `utf8`"
        ));
    }

    function _destruct() {
        $this->_conn = null;
    }

    /**
     * $opts = (
     *     'query' => string
     * )
     */
    function read($dsn) {
        $this->getinfo($dsn);
        $query = $this->opts['query'] ?
            "SELECT * FROM {$this->type} WHERE ".$this->opts['query'] :
            "SELECT * FROM {$this->type} LIMIT 10";
        $doc = new DOMDocument();
        $doc->preserveWhiteSpace = false;
        $doc->encoding = 'UTF-8';
        $doc->formatOutput = true;
        $result = $doc->appendChild($doc->createElement('result'));
        $result->setAttribute('dsn', $this->dsn);

        $stmt = $this->_conn->prepare($query);
        $stmt->execute();

        $this->_createnode($result, 'code', 0);
        $this->_createnode($result, 'message', 'OK');
        $rows = $this->_createnode($result, 'rows');
        $rows->setAttribute("table", $this->type);
        while ($rset = $stmt->fetch(PDO::FETCH_ASSOC)) {
            $row = $this->_createnode($rows, 'row');
            foreach ($rset as $key => $val) {
                $this->_createnode($row, $key, $val);
            }
        }

        return $doc;
    }
}

/**
 * media datasource
 */
class MediaDataSource extends DataSourceObject {
    function read($dsn) {
        $this->getinfo($dsn);
        $fname = mediaFN($this->type);
        if (!extension_loaded('zip'))
            return $this->error('1', 'NO ZIP EXTENSION');
        if (!@file_exists($fname))
            return $this->error('2', 'NOT EXISTS');
        $zh = zip_open($fname);
        if (!is_resource($zh))
            return $this->error($zh, 'ZIP ERROR');

        //Read entries and sort by lowcase/upcase complete.
        $entries = array();
        while ($e = zip_read($zh)) {
            $entries[zip_entry_name($e)] = $e;
        }
        ksort($entries, SORT_STRING);

        //Read tree.
        $tree = array();
        foreach ($entries as $e) {
            $curr = &$tree;
            $names = explode('/', zip_entry_name($e));
            for ($i = 0; $i < count($names) - 1; $i++) {
                $name = $names[$i];
                //dbglog(zip_entry_name($e)."=>".$name.print_r($curr,true));
                if ($curr[$name]) {
                    $curr = &$curr[$name];
                } else {
                    $curr[$name] = array();
                    $curr = &$curr[$name];
                }
            }
            if (strlen($names[count($names) - 1]) > 0) {
                $curr[$names[count($names) - 1]] = $e;
            }
        }

        //create DOM
        $doc = new DOMDocument('1.0', 'UTF-8');
        $doc->preserveWhiteSpace = false;
        $doc->formatOutput = true;
        $result = $doc->appendChild($doc->createElement('result'));
        $result->setAttribute("dsn", $this->dsn);
        $result->appendChild($doc->createElement('code', 0));
        $result->appendChild($doc->createElement('message', 'OK'));
        $zip = $result->appendChild($doc->createElement('zip'));
        $this->makedom($zip, $tree, '');

        zip_close($zh);
        return $doc;
    }

    function makedom(&$node, $tree, $partname) {
        $doc = $node->ownerDocument;
        if (is_resource($tree)) {
            $entry = $node->appendChild($doc->createElement('entry'));
            $entry->setAttribute('name', $partname);
            $entry->appendChild($doc->createElement('longname',
                    zip_entry_name($tree)));
            $entry->appendChild($doc->createElement('filesize',
                    zip_entry_filesize($tree)));
            $entry->appendChild($doc->createElement('compressedsize',
                    zip_entry_compressedsize($tree)));
            $entry->appendChild($doc->createElement('compressionmethod',
                    zip_entry_compressionmethod($tree)));
        } else if (is_array($tree)) {
            $entries = $node->appendChild($doc->createElement('entries'));
            $entries->setAttribute('name', $partname);
            //$entries->appendChild($doc->createElement('name', $partname));
            foreach ($tree as $name => $child) {
                $this->makedom($entries, $child, $name);
            }
        } else {
            dbglog("error");
        }
    }

    function error($code, $message) {
        $doc = new DOMDocument('1.0', 'UTF-8');
        $doc->preserveWhiteSpace = false;
        $doc->formatOutput = true;
        $result = $doc->appendChild($doc->createElement('result'));
        $result->appendChild($doc->createElement('code', $code));
        $result->appendChild($doc->createElement('message', $message));
        return $doc;
    }
}
